import re
from nltk import word_tokenize
from typing import Set, Dict


def clean_text(
    text: str,
    keep_pattern: str = "[a-zA-Z0-9!.,?]",
    drop_patterns: Set[str]=set([]),
    replace: Dict = {},
    maxlen: int = 15,
) -> str:
    """Tokenize and clean text, by matching it against keep_pattern and droping and replacing provided patterns.

    Args:
        text (str): Input text.
        keep_pattern (str): Allowed patterns e.g. [a-zA-Z]. Defaults to "[a-zA-Z0-9!.,?]". 
        drop_patterns (set): Set of patterns that should be dropeed from text.
        replace (dict): Dictionary of to_be_replaced_pattern: replaced_with. E.g. {[0-9]+: NUM}
        maxlen (int): Maximum length of a token. Defaults to 15. 

    Returns:
        out_text (string): Tokenized and cleaned up text with respect to all above criteria.
    """
    if not isinstance(text, str):
        raise TypeError("Input must be a string.")
    if len(text) == 0:
        raise ValueError("Input must be a non empty string.")

    # Drop unwanted tokens: Replace them with space, then replace resulting \s{2, } with one space
    for d in drop_patterns:  # d = e.g. <br/>
        if re.search(d, text):
            text = re.sub(d, " ", text)
    text = re.sub("\s{2,}", " ", text)

    for k, v in replace.items():
        if re.search(k, text):
            text = re.sub(k, v, text)

    tokens = word_tokenize(text)
    out_tokens = []
    for t in tokens:
        if len(t) < maxlen:
            if re.match(keep_pattern, t):
                out_tokens.append(t)
    out_text = " ".join(out_tokens)
    return out_text



from typing import Dict, List
import json
import sys
import pandas
import re
import tqdm
from snlp import logger
import nltk
from collections import Counter


def replace_mwes(
    path_to_mwes: str,
    mwe_types: List[str],
    df: pandas.DataFrame,
    text_column: str,
    am_threshold: float = 0.7,
    only_mwes: bool = False,
    lower_case: bool = False,
) -> pandas.DataFrame:
    """Hyphenates the mwes in the corpus so that they are treated as a single token by downstream applications.

    Args:
        path_to_mwes: Path to a json file that contains a dictionary of MWE type for each type,
                    unique MWEs to their count. E.g. {'NC': {'mwe1': 10}}.
        mwe_types: Types of MWEs to be replaced. Can be any of [NC, JNC].
        df: DataFrame comprising training data with a tokenized text column.
        text_column: Text (content) column of df.
        am_threshold: MWEs with an am greater than or equal to this threshold are selected for replacement.
        only_mwes: Whether or not keep only MWEs and drop the rest of the text.
        lower_case: Whether or not lowercase the sentence before replacing MWEs.

    Returns:
        df (pandas.FataFrame)
    """
    try:
        with open(path_to_mwes, "r") as file:
            mwe_type_mwe_am = json.load(file)
    except Exception as e:
        raise e
    good_mwes = set()
    for t in mwe_types:
        logger.info(f'Replacing MWEs of type {t} in the corpus.')
        if t not in mwe_type_mwe_am:
            logger.error(f'MWEs of type {t} do not exist. Make sure the file containing MWEs in {path_to_mwes} includes type {t}')
            sys.exit(1)
        pmi_sorted_dict = mwe_type_mwe_am[t]
        logger.info(f"Number of all MWEs of type {t}: {len(pmi_sorted_dict)}")
        for k, v in pmi_sorted_dict.items():
            if v >= am_threshold:
                good_mwes.add(k)
            else:
                break
        logger.info("Number of MWEs to be replaced in corpus based on the association threshold: %d" % len(good_mwes))

    logger.info("Replacing compounds in text")
    new_text = []
    for sent in tqdm.tqdm(df[text_column]):
        sent = sent.lower() if lower_case else sent
        bigrams = get_ngrams(sent, 2)
        if only_mwes:
            tmp = ""
            for bg in bigrams:
                if bg in good_mwes:
                    tmp += bg.split(" ")[0] + "-" + bg.split(" ")[1] + " "
            if tmp != "":
                sent = tmp.strip()
        else:
            for bg in bigrams:
                if bg in good_mwes:
                    sent = re.sub(bg, bg.split(" ")[0] + "-" + bg.split(" ")[1], sent)
        new_text.append(sent)
    df[text_column] = new_text
    return df


def get_ngrams(sentence: str, n: int) -> List:
    """Extracts n-grams from sentence.

    Args:
        sentence: Input sentence from which n-grams are to be extracted.
        n: Size of n-grams.

    Returns:
        ngrams: List of extracted n-grams.
    """
    ngrams = []
    try:
        tokens = sentence.split(" ")
    except Exception as E:
        logger.error(E)
        logger.error(f"Input sentence {sentence} cannot be spilitted around space. No n-gram is extracted.")
        return ngrams
    for i in range(len(tokens) - n + 1):
        ngrams.append(" ".join(tokens[i : i + n]))
    return ngrams


def get_counts(df: pandas.DataFrame, text_column: str, mwe_types: List[str]) -> dict:
    """Read a corpus in pandas.DataFrame format and generates all counts necessary for calculating AMs.

    Args:
        df (pandas.DataFrame): DataFrame with input data, which contains a column with text content
                               from which compounds and their counts are extracted.
        text_column: Name of the column the contains the text content.
        mwe_types: Types of MWEs. Can be any of [NC, JNC]

    Returns:
        res: Dictionary of mwe_types to dictionary of individual mwe within that type and their count.
            E.g. {'NC':{'climate change': 10, 'brain drain': 3}, 'JNC': {'black sheep': 3, 'red flag': 2}}
    """
    res = {}
    for mt in mwe_types:
        res[mt] = {}
    res["WORDS"] = {}
    for sent in tqdm.tqdm(df[text_column]):
        tokens = sent.split(" ")
        word_count_dict = Counter(tokens)
        for k, v in word_count_dict.items():
            if k in res["WORDS"]:
                res["WORDS"][k] += v
            else:
                res["WORDS"][k] = v
        for mt in mwe_types:
            mwes_count_dic = extract_mwes_from_sent(tokens, mwe_type=mt)
            for k, v in mwes_count_dic.items():
                if k in res[mt]:
                    res[mt][k] += v
                else:
                    res[mt][k] = v
    return res


def extract_mwes_from_sent(tokens: List[str], mwe_type: str) -> Dict:
    """Extract two-word noun compounds from tokenized input.

    Args:
        tokens: A tokenized sentence, i.e. list of tokens.
        type: Type of MWE. Any of ['NC', 'JNC'].

    Returns:
        mwes_count_dic: Dictionary of compounds to their count.
    """
    if not isinstance(tokens, list):
        raise TypeError(
            f'Input argument "tokens" must be a list of string. Currently it is of type {type(tokens)} \
            with a value of: {tokens}.'
        )
    if len(tokens) == 0:
        return
    mwes = []
    postag_tokens = nltk.pos_tag(tokens)
    w1_pos_tags = []
    w2_pos_tags = []
    if mwe_type == "NC":
        w1_pos_tags = ["NN", "NNS"]
        w2_pos_tags = ["NN", "NNS"]
    elif mwe_type == "JNC":
        w1_pos_tags = ["JJ"]
        w2_pos_tags = ["NN", "NNS"]
    for i in range(len(postag_tokens) - 1):
        w1 = postag_tokens[i]
        if w1[1] not in w1_pos_tags:
            continue
        else:
            w2 = postag_tokens[i + 1]
            if not re.match("[a-zA-Z0-9]{2,}", w1[0]) or not re.match("[a-zA-Z0-9]{2,}", w2[0]):
                continue
            if w2[1] in w2_pos_tags:
                if i + 2 < len(postag_tokens):
                    w3 = postag_tokens[i + 2]
                    if w3 not in ["NN", "NNS"]:
                        mwes.append(w1[0] + " " + w2[0])
                else:
                    mwes.append(w1[0] + " " + w2[0])
    mwes_count_dic = Counter(mwes)
    return mwes_count_dic



import pandas as pd
import numpy as np
import itertools


from sklearn.metrics import confusion_matrix
from matplotlib import pyplot as plt
from collections import namedtuple
from sklearn.metrics import precision_score, recall_score


def measure_accuracy(data_frame):
    data_frame['right_prediction'] = data_frame.apply(right_prediction, axis=1)
    num_correct_predictions = data_frame.right_prediction.sum()
    accuracy = float(num_correct_predictions)/data_frame.shape[0]
    return accuracy, num_correct_predictions


def right_prediction(row):
        if row['label'] == row['prediction']:
            return 1
        else:
            return 0


def plot_confusion_matrix(cm, classes,
                          normalize=False,
                          title='Confusion matrix',
                          cmap=plt.cm.Blues):
    """
    This function prints and plots the confusion matrix.
    Normalization can be applied by setting `normalize=True`.
    """
    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
        print("Normalized confusion matrix")
    else:
        print('Confusion matrix, without normalization')

    plt.imshow(cm, interpolation='nearest', cmap=cmap)
    plt.title(title)
    plt.colorbar()
    tick_marks = np.arange(len(classes))
    plt.xticks(tick_marks, classes, rotation=90)
    plt.yticks(tick_marks, classes)

    fmt = '.2f' if normalize else 'd'
    thresh = cm.max() / 2.
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        plt.text(j, i, format(cm[i, j], fmt),
                 horizontalalignment="center",
                 color="white" if cm[i, j] > thresh else "black")

    plt.tight_layout()
    plt.ylabel('True label')
    plt.xlabel('Predicted label')


def evaluate_model(predictions, groundtruth, confustion_matrix_path=False):

    eval_df = pd.DataFrame(columns=['prediction', 'label'])
    eval_df['prediction'] = predictions
    eval_df['label'] = groundtruth

    accuracy, num_correct_predictions = measure_accuracy(eval_df)

    eval_df.sort_values(by=['label'], ascending=[True], inplace=True)
    
    if confustion_matrix_path:
        cnf_matrix = confusion_matrix(eval_df['label'], eval_df['prediction'])
        np.set_printoptions(precision=2)

        plt.figure(figsize=(30, 20))
        plt.rc('font', size=18)
        plot_confusion_matrix(cnf_matrix, classes=eval_df['label'].unique(), normalize=False,
                            title='Confusion matrix')
        plt.savefig(confustion_matrix_path, bbox_inches='tight')

    EvalResults = namedtuple('EvalResults', 'macro_p micro_p macro_rec micro_rec accuracy num_correct_predictions confusion_plot')
    return EvalResults(micro_p=precision_score(eval_df['label'], eval_df['prediction'], average='micro'),
                       macro_p=precision_score(eval_df['label'], eval_df['prediction'], average='macro'),
                       micro_rec=recall_score(eval_df['label'], eval_df['prediction'], average='micro'),
                       macro_rec=recall_score(eval_df['label'], eval_df['prediction'], average='macro'),
                       accuracy=accuracy,
                       num_correct_predictions=num_correct_predictions, 
                       confusion_plot=plt)



import os
import sys
import pandas
import json
from pathlib import Path
from typing import List
from nltk import word_tokenize
from snlp.mwes.am import calculate_am
from snlp.mwes.mwe_utils import replace_mwes, get_counts
from snlp import logger


class MWE(object):
    def __init__(
        self,
        df: pandas.DataFrame,
        text_column: str,
        mwe_types: List[str] = ["NC"],
        output_dir: str = "tmp",
        tokenize=False,
    ) -> None:
        """Provide functionalities around MWEs, for unsupervised extraction of MWEs from text and replacing
        them in the corpus.

        Args:
            df: DataFrame with a text_column that contains the corpus.
            text_col: Specifies the column of DataFrame that contains the corpus. 'text_column' must contain tokenized text.
            mwe_types: Types of MWEs. Can be a list containing any of ['NC', 'JNC'].
            output_dir: Output directory where counts, MWEs and corpus with replaced MWEs are stored.
            count_dir: Directory where count_file is sotred.
            count_file: File in which counts are sotred.
            mwe_dir: Directory where mwe_file is sotred.
            mwe_file: File in which MWEs are sotred.
            tokenize: Tokenize the content of 'text_column'.

        Returns:
            None
        """
        self.df = df
        self.text_col = text_column
        for mt in mwe_types:
            if mt not in ["NC", "JNC"]:
                raise ValueError(f"{mt} type is not recognized.")
        self.mwe_types = mwe_types

        self.output_dir = output_dir
        self.count_dir = os.path.join(self.output_dir, "counts")
        self.count_file = os.path.join(self.count_dir, "count_data.json")
        self.mwe_dir = os.path.join(self.output_dir, "mwes")
        self.mwe_file = os.path.join(self.mwe_dir, "mwe_data.json")
        
        Path(self.output_dir).mkdir(exist_ok=True)

        if tokenize:
            logger.info('"tokenize" flag set to True. This might lead to a slow instantiation.')
            self.df[text_column] = self.df[text_column].apply(self._tokenize)
        else:
            self._check_tokenized()

    def _tokenize(self, x):
        """Helper function to tokenize and join the results with a space.

        Args:
            None

        Returns:
            None
        """
        return " ".join(word_tokenize(x))

    def _check_tokenized(self) -> None:
        """Helper function to check if the content of text_column is tokenized.

        Args:
            None

        Returns:
            None
        """
        if self.df[self.text_col].shape[0] > 200:
            tests = self.df[self.text_col].sample(n=200).tolist()
        else:
            tests = self.df[self.text_col].sample(frac=0.8).tolist()
        num_pass = 0
        for t in tests:
            if " ".join(word_tokenize(t)) == t:
                num_pass += 1
        if float(num_pass) / float(len(tests)) < 0.8:
            logger.warning(
                f"It seems that the content of {self.text_col} in the input data frame is not (fully) tokenized.\nThis can lead to poor results. Consider re-instantiating your MWE instance with 'tokenize' flag set to True.\nNote that this might lead to a slower instantiation."
            )

    def build_counts(self, file_name: str=None) -> None:
        """Create various count files to be used by downstream methods 
        by calling snlp.mwes.counter.get_counts.

        Args:
            None

        Returns:
            None
        """
        logger.info("Creating counts...")
        res = get_counts(df=self.df, text_column=self.text_col, mwe_types=self.mwe_types)
        # Directory
        try:
            Path(self.count_dir).mkdir(exist_ok=True)
        except Exception as e:
            logger.error(e)
            raise e
        # File
        if not file_name:
            try:
                with open(self.count_file, "w") as file:
                    json.dump(res, file)
            except Exception as e:
                logger.error(e)
                raise e
        else:
            try:            
                with open(file_name, "w") as file:
                    json.dump(res, file)
            except Exception as e:
                logger.error(e)
                raise e
            self.count_file = file_name

    def extract_mwes(self, am: str = "pmi", file_name: str=None) -> None:
        """
        Args:
            mwe_types: Types of MWEs. Can be any of [NC, JNC]
            am: The association measure to be used. Can be any of [pmi, npmi]

        Returns:
            None
        """
        with open(self.count_file, "r") as file:
            count_data = json.load(file)
        logger.info(f"Extracting {self.mwe_types} based on {am}")
        mwe_am_dict = calculate_am(count_data=count_data, am=am, mwe_types=self.mwe_types)
        # Dir
        try:
            Path(self.mwe_dir).mkdir(exist_ok=True)
        except Exception as e:
            logger.error(e)
            raise e
        # File
        if not file_name:
            try:
                with open(self.mwe_file, "w") as file:
                    json.dump(mwe_am_dict, file)
            except Exception as e:
                logger.error(e)
                raise e
        else:
            try:
                with open(file_name, "w") as file:
                    json.dump(mwe_am_dict, file)
            except Exception as e:
                logger.error(e)
                raise e


if __name__ == "__main__":
    import pandas as pd
    from snlp.mwes.mwe_utils import replace_mwes

    imdb_train = pd.read_csv("data/imdb_train_sample.tsv", sep="\t", names=["label", "text"])
    mwe = MWE(df=imdb_train, mwe_types=["NC", "JNC"], text_column="text", tokenize=True)
    mwe.build_counts()
    mwe.extract_mwes(am="npmi")
    new_df = replace_mwes(
        path_to_mwes="tmp/mwes/mwe_data.json", mwe_types=["NC", "JNC"], df=imdb_train, text_column="text"
    )
    new_df.to_csv("tmp/new_df.csv", sep="\t")



import unittest
from snlp.mwes.am import extract_ncs_from_sent, calculate_am, get_counts


class TestAms(unittest.TestCase):

    def test_extract_nsc(self):
        self.assertRaises(TypeError, extract_ncs_from_sent, 5)
        self.assertRaises(TypeError, extract_ncs_from_sent, -2)
        self.assertRaises(TypeError, extract_ncs_from_sent, [1, 2, 3])
        self.assertRaises(TypeError, extract_ncs_from_sent, [1, 2, 'test'])
        self.assertRaises(ValueError, extract_ncs_from_sent, '')

if __name__ == '__main__':
    unittest.main()


import pandas as pd
import plotly.express as px
from scipy.stats import zscore
from snlp import logger
from snlp import gaussianize
from sklearn.feature_extraction.text import TfidfVectorizer
from typing import Set, Dict


class RedunWords(object):
    def __init__(self, documents, method="idf"):
        """"Class to identify and represent a set of redundant words.

        Args:
            documents: An iterable which yields str.
            method: Method of creating the set of redundant words. Can be either of ['idf'].

        Returns:
            None
        """
        self.documents = documents
        self.method = method
    
    def _get_scores(self) -> Dict:
        """Helper method to generate statistical scores using the technique specified by self.method arg. 

        Args:
            None

        Returns:
            token_score_dict: Dictionary of tokens to their score
        """
        token_score_dict = {}
        if self.method == 'idf':
            vectorizer = TfidfVectorizer(min_df=1)
            X = vectorizer.fit_transform(self.documents)
            idf = vectorizer.idf_
            token_score_dict = dict(zip(vectorizer.get_feature_names(), idf))
        else:
            raise ValueError(f'Currently, the only available method is idf but you have specified {self.method}')
        return token_score_dict

    def get_redundant_terms(self, z=3, manual=False, manual_thresholds: dict={'lower_threshold':-1, 'upper_threshold': -1}):
        """Create a filter set by identifying words with anomalous statistics.

        Args:
            documents: An iterable which yields either str, unicode or file objects.
            manual: Whether or not select redundant words using manual thresholds.
            z: Words with a Z-score above this value are considered redundant. Default is 3.
            l_idf: Lower cut-off threshold for IDF (inclusive). Used only when 
            u_idf: Upper cut-off threshold for IDF (inclusive).

        Returns:
            redundant_word_set (set): Set of redundant words.
        """
        token_score_dict = self._get_scores()
        if not manual:
            redundant_word_set = self._redundant_terms_zscore(token_score_dict=token_score_dict, z_value=z)
        else:
            redundant_word_set = self._redundant_terms(token_score_dict=token_score_dict,
                                            lower_threshold=manual_thresholds['lower_threshold'],
                                            upper_threshold=manual_thresholds['upper_threshold'])
        return redundant_word_set

    def _redundant_terms(self, token_score_dict, lower_threshold, upper_threshold) -> Set[str]:
        """Identify redundant terms by looking at manual thresholds i.e. lower_threshold, upper_threshold. 
        
        Args:
            token_metric_dict: Dictionary of tokens to their metric.
            lower_threshold: Lower cut-off threshold (inclusive).
            upper_threshold: Upper cut-off threshold (inclusive).

        Returns:
            redundant_terms: Set of redundant terms.
        """
        sl = sorted(token_score_dict.items(), key=lambda kv: kv[1])
        smallest_idf = sl[0][1]
        largest_idf = sl[len(sl) - 1][1]
        if lower_threshold < smallest_idf:
            raise ValueError(
                f'Idf values are between {smallest_idf:.2f} and {largest_idf:.2f}. You have set the lower_idf to {lower_threshold:.2f}. Update the values accordingly, \
                or consider switching the manual flag back to False.'
            )
        if upper_threshold > largest_idf:
            raise ValueError(
                f'Idf values are between {smallest_idf:.2f} and {largest_idf:.2f}. You have set the upper_idf to {upper_threshold:.2f}. Update the values accordingly, \
                or consider switching the manual flag back to False.'
            )
        redundant_terms = set()
        for i in range(len(sl)):
            if sl[i][1] < lower_threshold:
                redundant_terms.add(sl[i][0])
            if sl[i][1] >= upper_threshold:
                redundant_terms.add(sl[i][0])
        return redundant_terms

    def _redundant_terms_zscore(self, token_score_dict, z_value):
        """Identify redundant terms by looking at the provided Z-score.

        Args:
            token_metric_dict: Dictionary of tokens to their metric.
            z_value: Words with a Z-score above this value are considered redundant.

        Returns:
            redundant_terms: Set of redundant terms.
        """
        idf_df = pd.DataFrame(token_score_dict.items(), columns=["token", "idf"])
        # Gaussianize idf
        g = gaussianize.Gaussianize(strategy="brute")
        g.fit(idf_df["idf"])
        idf_guassian = g.transform(idf_df["idf"])
        idf_df["idf_gaussianized"] = idf_guassian
        # Calculate z score
        z = zscore(idf_df["idf_gaussianized"])
        idf_df["zscore"] = z
        redundant_terms = set()
        for i in range(len(idf_df)):
            if idf_df.iloc[i]["zscore"] <= -z_value or idf_df.iloc[i]["zscore"] >= z_value:
                redundant_terms.add(idf_df.iloc[i]["token"])
        return redundant_terms

    def show_plot(self) -> None:
        """ Create a histogram for the scores to help identify the cut-off threshold.

        Args:
            scores: 
        Returns:
            None
        """
        token_score_dict = self._get_scores()
        scores = [v for k,v in token_score_dict.items()]
        # l = len(scores)
        # print(l)
        # print(scores[l-100:l])
        score_name = f'{self.method} scores'
        scores_df = pd.DataFrame(scores, columns=[score_name])
        fig = px.histogram(scores_df, x=score_name, marginal='rug', color_discrete_sequence=['magenta'])
        fig.show()


if __name__ == "__main__":
    imdb_train = pd.read_csv('resources/data/imdb_train_sample.tsv', sep='\t', names=['label', 'text'])
    rw = RedunWords(imdb_train["text"])
    rw.show_plot()
    red_words = rw.get_redundant_terms()
    print(red_words)


# def filter_text(text, filter_set):
#     """
#     Args:
#         text (list): tokenized text
#         filterset (set): Set of filter words
#     """
#     if not isinstance(text, list):
#         raise TypeError("Input must be a list.")
#     if len(text) == 0:
#         raise ValueError("Input must be a non empty list.")
#     res = " ".join([t for t in text if t not in filter_set])
#     return res



# def save_filterset_tofile(filter_set, path):
#     with open(path, "w") as f:
#         for word in filter_set:
#             f.write(word + "\n")
#     f.close()


# def create_filterset_map(p2_raw_train_df, output_dir, zs=[3]):
#     original_df = pd.read_csv(p2_raw_train_df, sep="\t", names=["label", "text"])
#     wf = RedunWords()
#     z_map = {}
#     for z in zs:
#         filter_words = wf.get_redundant_terms(original_df.text, method="automatic", z=z)
#         save_filterset_tofile(filter_words, os.path.join(output_dir, "z" + str(z) + "_filterset.txt"))
#         z_map[z] = filter_words
#     return z_map


# def create_filterset(p2_raw_train_df, output_dir):
#     original_df = pd.read_csv(p2_raw_train_df, sep="\t", names=["label", "text"])
#     wf = RedunWords()
#     filter_words = wf.get_redundant_terms(original_df.text, method="manual", z=None)
#     save_filterset_tofile(filter_words, os.path.join(output_dir, "manual_filterset.txt"))
#     return filter_words



"""
Transform data so that it is approximately normally distributed

This code written by Greg Ver Steeg, 2015.
"""

import numpy as np
from scipy.special import lambertw
from scipy.stats import kurtosis, norm, rankdata, boxcox
from scipy.optimize import fmin  # TODO: Explore efficacy of other opt. methods
import sklearn

np.seterr(all='warn')


class Gaussianize(sklearn.base.TransformerMixin):
    """
    Gaussianize data using various methods.

    Conventions
    ----------
    This class is a wrapper that follows sklearn naming/style (e.g. fit(X) to train).
    In this code, x is the input, y is the output. But in the functions outside the class, I follow
    Georg's convention that Y is the input and X is the output (Gaussianized) data.

    Parameters
    ----------
    tol : float, default = 1e-4

    max_iter : int, default = 100
        Maximum number of iterations to search for correct parameters of Lambert transform.

    strategy : str, default='lambert'
        Possibilities are 'lambert'[1], 'brute'[2] and 'boxcox'[3].

    Attributes
    ----------
    coefs_ : list of tuples
        For each variable, we have transformation parameters.
        For Lambert, e.g., a tuple consisting of (mu, sigma, delta), corresponding to the parameters of the
        appropriate Lambert transform. Eq. 6 and 8 in the paper below.

    References
    ----------
    [1] Georg Goerg. The Lambert Way to Gaussianize heavy tailed data with
                        the inverse of Tukey's h transformation as a special case
        Author generously provides code in R: https://cran.r-project.org/web/packages/LambertW/
    [2] Valero Laparra, Gustavo Camps-Valls, and Jesus Malo. Iterative Gaussianization: From ICA to Random Rotations
    [3] Box cox transformation and references: https://en.wikipedia.org/wiki/Power_transform
    """

    def __init__(self, tol=1.22e-4, max_iter=100, verbose=False, strategy='lambert'):
        self.tol = tol
        self.max_iter = max_iter
        self.strategy = strategy
        self.coefs_ = []  # Store tau for each transformed variable
        self.verbose = verbose

    def fit(self, x, y=None):
        """Fit a Gaussianizing transformation to each variable/column in x."""
        x = np.asarray(x)
        if len(x.shape) == 1:
            x = x[:, np.newaxis]
        elif len(x.shape) != 2:
            print("Data should be a 1-d list of samples to transform or a 2d array with samples as rows.")

        if self.strategy == 'lambert':
            if self.verbose:
                print("Gaussianizing with Lambert method")
            for x_i in x.T:
                self.coefs_.append(igmm(x_i, tol=self.tol, max_iter=self.max_iter))
        elif self.strategy == 'brute':
            for x_i in x.T:
                self.coefs_.append(None)  # TODO: In principle, we could store parameters to do a quasi-invert
        elif self.strategy == 'boxcox':
            for x_i in x.T:
                self.coefs_.append(boxcox(x_i)[1])
        else:
            raise NotImplementedError
        return self

    def transform(self, x):
        """Transform new data using a previously learned Gaussianization model."""
        x = np.asarray(x)
        if len(x.shape) == 1:
            x = x[:, np.newaxis]
        elif len(x.shape) != 2:
            print("Data should be a 1-d list of samples to transform or a 2d array with samples as rows.")
        if x.shape[1] != len(self.coefs_):
            print("%d variables in test data, but %d variables were in training data." % (x.shape[1], len(self.coefs_)))

        if self.strategy == 'lambert':
            return np.array([w_t(x_i, tau_i) for x_i, tau_i in zip(x.T, self.coefs_)]).T
        elif self.strategy == 'brute':
            return np.array([norm.ppf((rankdata(x_i) - 0.5) / len(x_i)) for x_i in x.T]).T
        elif self.strategy == 'boxcox':
            return np.array([boxcox(x_i, lmbda=lmbda_i) for x_i, lmbda_i in zip(x.T, self.coefs_)]).T
        else:
            raise NotImplementedError

    def inverse_transform(self, y):
        """Recover original data from Gaussianized data."""
        if self.strategy == 'lambert':
            return np.array([inverse(y_i, tau_i) for y_i, tau_i in zip(y.T, self.coefs_)]).T
        elif self.strategy == 'boxcox':
            return np.array([(1. + lmbda_i * y_i) ** (1./lmbda_i) for y_i, lmbda_i in zip(y.T, self.coefs_)]).T
        else:
            print('Inversion not supported for this gaussianization transform.')
            raise NotImplementedError

    def qqplot(self, x, prefix='qq'):
        """Show qq plots compared to normal before and after the transform."""
        from matplotlib import pylab
        from scipy.stats import probplot
        y = self.transform(x)

        for i, (x_i, y_i) in enumerate(zip(x.T, y.T)):
            probplot(x_i, dist="norm", plot=pylab)
            pylab.savefig(prefix + '_%d_before.png' % i)
            pylab.clf()

            probplot(y_i, dist="norm", plot=pylab)
            pylab.savefig(prefix + '_%d_after.png' % i)
            pylab.clf()


def w_d(z, delta):
    # Eq. 9
    if delta < 1e-6:
        return z
    return np.sign(z) * np.sqrt(np.real(lambertw(delta * z ** 2)) / delta)


def w_t(y, tau):
    # Eq. 8
    return tau[0] + tau[1] * w_d((y - tau[0]) / tau[1], tau[2])


def inverse(x, tau):
    # Eq. 6
    u = (x - tau[0]) / tau[1]
    return tau[0] + tau[1] * (u * np.exp(u * u * (tau[2] * 0.5)))


def igmm(y, tol=1.22e-4, max_iter=100):
    # Infer mu, sigma, delta using IGMM in Alg.2, Appendix C
    if np.std(y) < 1e-4:
        return np.mean(y), np.std(y).clip(1e-4), 0
    delta0 = delta_init(y)
    tau1 = (np.median(y), np.std(y) * (1. - 2. * delta0) ** 0.75, delta0)
    for k in range(max_iter):
        tau0 = tau1
        z = (y - tau1[0]) / tau1[1]
        delta1 = delta_gmm(z)
        x = tau0[0] + tau1[1] * w_d(z, delta1)
        mu1, sigma1 = np.mean(x), np.std(x)
        tau1 = (mu1, sigma1, delta1)

        if np.linalg.norm(np.array(tau1) - np.array(tau0)) < tol:
            break
        else:
            if k == max_iter - 1:
                print("Warning: No convergence after %d iterations. Increase max_iter." % max_iter)
    return tau1


def delta_gmm(z):
    # Alg. 1, Appendix C
    delta0 = delta_init(z)

    def func(q):
        u = w_d(z, np.exp(q))
        if not np.all(np.isfinite(u)):
            return 0.
        else:
            k = kurtosis(u, fisher=True, bias=False)**2
            if not np.isfinite(k) or k > 1e10:
                return 1e10
            else:
                return k

    res = fmin(func, np.log(delta0), disp=0)
    return np.around(np.exp(res[-1]), 6)


def delta_init(z):
    gamma = kurtosis(z, fisher=False, bias=False)
    with np.errstate(all='ignore'):
        delta0 = np.clip(1. / 66 * (np.sqrt(66 * gamma - 162.) - 6.), 0.01, 0.48)
    if not np.isfinite(delta0):
        delta0 = 0.01
    return delta0


if __name__ == '__main__':
    # Command line interface
    # Sample commands:
    # python gaussianize.py test_data.csv
    import csv
    import sys, os
    import traceback
    from optparse import OptionParser, OptionGroup

    parser = OptionParser(usage="usage: %prog [options] data_file.csv \n"
                                "It is assumed that the first row and first column of the data CSV file are labels.\n"
                                "Use options to indicate otherwise.")
    group = OptionGroup(parser, "Input Data Format Options")
    group.add_option("-c", "--no_column_names",
                     action="store_true", dest="nc", default=False,
                     help="We assume the top row is variable names for each column. "
                          "This flag says that data starts on the first row and gives a "
                          "default numbering scheme to the variables (1,2,3...).")
    group.add_option("-r", "--no_row_names",
                     action="store_true", dest="nr", default=False,
                     help="We assume the first column is a label or index for each sample. "
                          "This flag says that data starts on the first column.")
    group.add_option("-d", "--delimiter",
                     action="store", dest="delimiter", type="string", default=",",
                     help="Separator between entries in the data, default is ','.")
    parser.add_option_group(group)

    group = OptionGroup(parser, "Transform Options")
    group.add_option("-s", "--strategy",
                     action="store", dest="strategy", type="string", default="lambert",
                     help="Strategy.")
    parser.add_option_group(group)

    group = OptionGroup(parser, "Output Options")
    group.add_option("-o", "--output",
                     action="store", dest="output", type="string", default="gaussian_output.csv",
                     help="Where to store gaussianized data.")
    group.add_option("-q", "--qqplots",
                     action="store_true", dest="q", default=False,
                     help="Produce qq plots for each variable before and after transform.")
    parser.add_option_group(group)

    (options, args) = parser.parse_args()
    if not len(args) == 1:
        print("Run with '-h' option for usage help.")
        sys.exit()

    #Load data from csv file
    filename = args[0]
    with open(filename, 'rU') as csvfile:
        reader = csv.reader(csvfile, delimiter=" ")  #options.delimiter)
        if options.nc:
            variable_names = None
        else:
            variable_names = reader.next()[(1 - options.nr):]
        sample_names = []
        data = []
        for row in reader:
            if options.nr:
                sample_names = None
            else:
                sample_names.append(row[0])
            data.append(row[(1 - options.nr):])

    print(len(data), data[0])
    try:
        for i in range(len(data)):
            data[i] = map(float, data[i])
        X = np.array(data, dtype=float)  # Data matrix in numpy format
    except:
        print("Incorrect data format.\nCheck that you've correctly specified options " \
              "such as continuous or not, \nand if there is a header row or column.\n" \
              "Run 'python gaussianize.py -h' option for help with options.")
        traceback.print_exc(file=sys.stdout)
        sys.exit()

    ks = []
    for xi in X.T:
        ks.append(kurtosis(xi))
    print(np.mean(np.array(ks) > 1))
    from matplotlib import pylab
    pylab.hist(ks, bins=30)
    pylab.xlabel('excess kurtosis')
    pylab.savefig('excess_kurtoses_all.png')
    pylab.clf()
    pylab.hist([k for k in ks if k < 2], bins=30)
    pylab.xlabel('excess kurtosis')
    pylab.savefig('excess_kurtoses_near_zero.png')
    print(np.argmax(ks))
    pdict = {}
    for k in np.argsort(- np.array(ks))[:50]:
        pylab.clf()
        p = np.argmax(X[:, k])
        pdict[p] = pdict.get(p, 0) + 1
        pylab.hist(X[:, k], bins=30)
        pylab.xlabel(variable_names[k])
        pylab.ylabel('Histogram of patients')
        pylab.savefig('high_kurtosis/'+variable_names[k] + '.png')
    print(str(pdict)+"# 203, 140 appear three times.")
    sys.exit()
    out = Gaussianize(strategy=options.strategy)
    y = out.fit_transform(X)
    with open(options.output, 'w') as csvfile:
        writer = csv.writer(csvfile, delimiter=options.delimiter)
        if not options.nc:
            writer.writerow([""] * (1 - options.nr) + variable_names)
        for i, row in enumerate(y):
            if not options.nr:
                writer.writerow([sample_names[i]] + list(row))
            else:
                writer.writerow(row)

    if options.q:
        print('Making qq plots')
        prefix = options.output.split('.')[0]
        if not os.path.exists(prefix+'_q'):
            os.makedirs(prefix+'_q')
        out.qqplot(X, prefix=prefix + '_q/q')


import string
from typing import List, Tuple
import nltk
import os
import plotly
import pandas
import plotly.figure_factory as ff
import plotly.graph_objs as go

from wordcloud import WordCloud, get_single_color_func
from snlp import logger
from tqdm import tqdm
from nltk.corpus import stopwords
from plotly.subplots import make_subplots


def create_adjust_subplots(labels: List[Tuple]) -> plotly.graph_objects.Figure:
    """Create subplots and adjust the location of the titles wrt the number of labels.

    Args:
        labels (list): List of (label, type) tuples.

    Returns:
        fig (plotly.graph_objs.Figure)
    """
    if len(labels) == 0:
        titles = (
            "Analysis of Text",
            "Analysis of Labels",
            "Document Lengths",
            "",
            "",
            "",
            "",
            "Word Frequency",
            "",
            "",
            "",
            "",
            "" "Common Nouns",
            "",
            "",
            "",
            "Common Adjectives",
            "",
            "",
            "",
            "Common Verbs",
        )
    elif len(labels) == 1:
        titles = (
            "Analysis of Text",
            "Analysis of Labels",
            "Document Lengths",
            labels[0][0].capitalize(),
            "",
            "",
            "Word Frequency",
            "",
            "",
            "",
            "",
            "Common Nouns",
            "",
            "",
            "",
            "Common Adjectives",
            "",
            "",
            "",
            "Common Verbs",
        )
    elif len(labels) == 2:
        titles = (
            "Analysis of Text",
            "Analysis of Labels",
            "Document Lengths",
            labels[0][0].capitalize(),
            "",
            labels[1][0].capitalize(),
            "Word Frequency",
            "",
            "",
            "Common Nouns",
            "",
            "",
            "",
            "Common Adjectives",
            "",
            "",
            "",
            "Common Verbs",
        )
    elif len(labels) == 3:
        titles = (
            "Analysis of Text",
            "Analysis of Labels",
            "Document Lengths",
            labels[0][0].capitalize(),
            "",
            labels[1][0].capitalize(),
            "Word Frequency",
            labels[2][0].capitalize(),
            "",
            "Common Nouns",
            "",
            "",
            "",
            "Common Adjectives",
            "",
            "",
            "",
            "Common Verbs",
        )
    elif len(labels) == 4:
        titles = (
            "Analysis of Text",
            "Analysis of Labels",
            "Document Lengths",
            labels[0][0].capitalize(),
            "",
            labels[1][0].capitalize(),
            "Word Frequency",
            labels[2][0].capitalize(),
            "",
            "Common Nouns",
            labels[3][0].capitalize(),
            "",
            "Common Adjectives",
            "",
            "",
            "",
            "Common Verbs",
        )
    fig = make_subplots(
        rows=16,
        cols=2,
        subplot_titles=titles,
        specs=[
            [{}, {}],
            [{"rowspan": 2}, {"rowspan": 2} if len(labels) >= 1 else {}],  # row 2
            [None, None if len(labels) >= 1 else {}],
            [{}, {"rowspan": 2} if len(labels) >= 2 else {}],  # row 4
            [{"rowspan": 2}, None if len(labels) >= 2 else {}],
            [None, {"rowspan": 2} if len(labels) >= 3 else {}],  # row 6
            [{}, None if len(labels) >= 3 else {}],
            [{"rowspan": 3}, {"rowspan": 2} if len(labels) >= 4 else {}],  # row 8
            [None, None if len(labels) >= 4 else {}],
            [None, {}],
            [{"rowspan": 3}, {}],  # 12
            [None, {}],
            [None, {}],
            [{"rowspan": 3}, {}],  # 16
            [None, {}],
            [None, {}],
        ],
        vertical_spacing=0.035,
    )
    return fig


def plotly_wordcloud(token_count_dic: dict) -> plotly.graph_objects.Scatter:
    """Create a world cloud trace for plotly.

    Args:
        token_count_dic (dictionary): Dictionary of token to its count

    Returns:
        trace (plotly.graph_objects.Scatter)
    """
    wc = WordCloud(color_func=get_single_color_func("deepskyblue"), max_words=100)
    wc.generate_from_frequencies(token_count_dic)

    word_list = []
    rel_freq_list = []
    freq_list = []
    fontsize_list = []
    position_list = []
    orientation_list = []
    color_list = []

    for (word, rel_freq), fontsize, position, orientation, color in wc.layout_:
        word_list.append(word)
        rel_freq_list.append(rel_freq)
        freq_list.append(token_count_dic[word])
        fontsize_list.append(fontsize)
        position_list.append(position)
        orientation_list.append(orientation)
        color_list.append(color)

    # get the positions
    x = []
    y = []
    for i in position_list:
        x.append(i[0])
        y.append(i[1])

    # get the relative occurence frequencies
    new_freq_list = []
    for i in rel_freq_list:
        new_freq_list.append(i * 100)
    new_freq_list

    trace = go.Scatter(
        x=x,
        y=y,
        textfont=dict(size=new_freq_list, color=color_list),
        hoverinfo="text",
        hovertext=["{0}: {1}".format(w, f) for w, f in zip(word_list, freq_list)],
        mode="text",
        text=word_list,
    )
    return trace


def generate_text_plots(
    figure: plotly.graph_objs.Figure,
    doc_length_list: List,
    word_freq_list: List,
    noun_freq_dict: dict,
    adj_freq_dic: dict,
    verb_freq_dic: dict,
) -> None:
    """Generate distribution plots and word clouds for the textual content based on the input argumnets.

    Args:
        figure (plotly.graph_objs.Figure): Figure object in which the plots are created.
        doc_length_list (list): List containing the length of each document.
        word_freq_list (list): List containing the frequency of each word.
        noun_freq_dict (dictionary): Dictionary of noun to frequency.
        adj_freq_dict (dictionary): Dictionary of noun to frequency.
        verb_freq_dict (dictionary): Dictionary of verb to frequency.

    Returns:
        None
    """
    # Customize plots and x, y labels: https://plotly.com/python/subplots/#customizing-subplot-axes
    # Word cloud with plotly: https://github.com/PrashantSaikia/Wordcloud-in-Plotly/blob/master/plotly_wordcloud.py
    def _distplot_to_dist_trace(data: List, color: str) -> dict:
        """Create a trace from data.
        Args:
            data (list): List of numbers
            color (list[str])

        Returns:
            data_dist (dict): Representation of a distplot figure.
        """
        data_dist = ff.create_distplot([data], group_labels=["distplot"], colors=[color])["data"]
        for item in data_dist:
            item.pop("xaxis", None)
            item.pop("yaxis", None)
        return data_dist

    data1_dist = _distplot_to_dist_trace(doc_length_list, color="rgb(0, 200, 200)")
    data2_dist = _distplot_to_dist_trace(word_freq_list, color="magenta")

    d1_hist = data1_dist[0]
    d1_kde = data1_dist[1]
    d1_rug = data1_dist[2]

    d2_hist = data2_dist[0]
    d2_kde = data2_dist[1]
    d2_rug = data2_dist[2]

    noun_cloud = plotly_wordcloud(token_count_dic=noun_freq_dict)
    adj_cloud = plotly_wordcloud(token_count_dic=adj_freq_dic)
    verb_cloud = plotly_wordcloud(token_count_dic=verb_freq_dic)

    figure.append_trace(d1_hist, 2, 1)
    figure.append_trace(d1_kde, 2, 1)
    figure.append_trace(d1_rug, 4, 1)

    figure.append_trace(d2_hist, 5, 1)
    figure.append_trace(d2_kde, 5, 1)
    figure.append_trace(d2_rug, 7, 1)

    figure.append_trace(noun_cloud, 8, 1)
    figure.append_trace(adj_cloud, 11, 1)
    figure.append_trace(verb_cloud, 14, 1)

    figure.update_xaxes(rangemode="tozero", row=4, col=1)
    figure.update_yaxes(showticklabels=False, row=4, col=1)

    figure.update_xaxes(rangemode="tozero", row=7, col=1)
    figure.update_yaxes(showticklabels=False, row=7, col=1)

    figure.update_yaxes(title_text="Probability", row=2, col=1)
    figure.update_yaxes(title_text="Probability", row=5, col=1)

    figure.update_xaxes(showticklabels=False, zeroline=False, row=8, col=1)
    figure.update_xaxes(showticklabels=False, zeroline=False, row=11, col=1)
    figure.update_xaxes(showticklabels=False, zeroline=False, row=14, col=1)
    figure.update_yaxes(showticklabels=False, zeroline=False, row=8, col=1)
    figure.update_yaxes(showticklabels=False, zeroline=False, row=11, col=1)
    figure.update_yaxes(showticklabels=False, zeroline=False, row=14, col=1)


def generate_label_plots(figure: plotly.graph_objs.Figure, df: pandas.DataFrame, label_cols: str) -> None:
    """Generate histogram and bar plots for the labels in label_cols.

    Args:
        figure (plotly.graph_objs.Figure): Figure object in which the plots are created.
        df (Pandas DataFrame): DataFrame that contains labels specified in label_cols.
        label_cols (list): list of tuples in the form of [('label_1', 'categorical/numerical'),
                           ('label_2', 'categorical/numerical'), ...]

    Returns:
        None
    """
    if len(label_cols) == 1:
        lab_trace1 = label_plot(df, label_col=label_cols[0][0], label_type=label_cols[0][1])
        figure.append_trace(lab_trace1, 2, 2)
        figure.update_yaxes(title_text="Count", row=2, col=2)
    elif len(label_cols) == 2:
        lab_trace1 = label_plot(df, label_col=label_cols[0][0], label_type=label_cols[0][1])
        lab_trace2 = label_plot(df, label_col=label_cols[1][0], label_type=label_cols[1][1])
        figure.append_trace(lab_trace1, 2, 2)
        figure.append_trace(lab_trace2, 4, 2)
        figure.update_yaxes(title_text="Count", row=2, col=2)
        figure.update_yaxes(title_text="Count", row=4, col=2)
    elif len(label_cols) == 3:
        lab_trace1 = label_plot(df, label_col=label_cols[0][0], label_type=label_cols[0][1])
        lab_trace2 = label_plot(df, label_col=label_cols[1][0], label_type=label_cols[1][1])
        lab_trace3 = label_plot(df, label_col=label_cols[2][0], label_type=label_cols[2][1])
        figure.append_trace(lab_trace1, 2, 2)
        figure.append_trace(lab_trace2, 4, 2)
        figure.append_trace(lab_trace3, 6, 2)
        figure.update_yaxes(title_text="Count", row=2, col=2)
        figure.update_yaxes(title_text="Count", row=4, col=2)
        figure.update_yaxes(title_text="Count", row=6, col=2)
    elif len(label_cols) == 4:
        lab_trace1 = label_plot(df, label_col=label_cols[0][0], label_type=label_cols[0][1])
        lab_trace2 = label_plot(df, label_col=label_cols[1][0], label_type=label_cols[1][1])
        lab_trace3 = label_plot(df, label_col=label_cols[2][0], label_type=label_cols[2][1])
        lab_trace4 = label_plot(df, label_col=label_cols[3][0], label_type=label_cols[3][1])
        figure.append_trace(lab_trace1, 2, 2)
        figure.append_trace(lab_trace2, 4, 2)
        figure.append_trace(lab_trace3, 6, 2)
        figure.append_trace(lab_trace4, 8, 2)
        figure.update_yaxes(title_text="Count", row=2, col=2)
        figure.update_yaxes(title_text="Count", row=4, col=2)
        figure.update_yaxes(title_text="Count", row=6, col=2)
        figure.update_yaxes(title_text="Count", row=8, col=2)


def label_plot(df: pandas.DataFrame, label_col: str, label_type: str) -> plotly.graph_objects.Histogram:
    """Create a plot for label_col in df, wrt to label_type.

    Args:
        df (Pandas DataFrame): DataFrame that contains label_col.
        label_col (str): Name of the label column in df that must be plotted.
        label_type (str): Represents the type of label and consequently specifies the type of plot.
                             It can be "numerical" or "categorical".

    Returns:
        trace (plotly.graph_objects.Histogram)
    """
    if label_type == "categorical":
        values = df[label_col].unique().tolist()  # ['pos', 'neg', 'neutral']
        counts = df[label_col].value_counts()  # 1212323
        x = []
        y = []
        for v in values:
            x.append(v)
            y.append(counts[v])
        trace = go.Bar(x=x, y=y, name=label_col)
    elif label_type == "numerical":
        trace = go.Histogram(x=df[label_col], name=label_col)
    else:
        raise ValueError('label_col input argument must be set to either "categorical" or "numerical".')
    return trace


def generate_report(
    df: pandas.DataFrame,
    out_dir: str,
    text_col: str,
    label_cols: List = [],
    language: str = "english",
    skip_stopwords_punc: bool = True,
    save_report: bool = False,
) -> None:
    """Generate analysis report and eitherr renders the report via Plotly show api or saves it offline to html.

    Args:
        df (pandas.DataFrame): DataFrame that contains text and labels.
        out_dir (str): Dir where the report is saved. Required only when save_report option is True.
        text_col (str): Name of the column that contains a tokenized text content.
        label_cols (list): list of tuples in the form of [('label_1', 'categorical/numerical'),
                           ('label_2', 'categorical/numerical'), ...]
        language (str): Language of the text in df[text_col]
        skip_stopwords_punc (bool): Whether or not skip stopwords and punctuations in the analysis. Default: True
        save_report (bool): Whether or not save the report as an html file. Default: False

    Returns:
        None
    """

    def update_count(items_dic: dict, items: List[str]) -> None:
        """Update the corresponding count for each key in  items_dic. w.r.t. terms in items.

        Args:
            items_dic (dict): Dictionary mapping keys to their count
            items (list): List of tokens

        Returns:
            None
        """
        for t in items:
            if t in items_dic:
                items_dic[t] += 1
            else:
                items_dic[t] = 1

    def get_pos(tagged_tokens: List[Tuple[str, str]], goal_pos: str) -> List:
        """Extracts goal_pos POS tags from tagged_tokens.

        Args:
            tagged_tokens (List[Tuple(str, str)]): Contains terms and ther pos tags. E.g.
                                                   [('cat', 'NN'), ('sat', 'VBD'), ('on', 'IN'), ('mat', 'NN')]
            goal_pos (str): Pos tag to look for in tagged_tokens

        Returns:
            res (List(str)): List of tokens with goal_pos pos tag
        """
        res = []
        for pt in tagged_tokens:
            if pt[1].startswith(goal_pos):
                res.append(pt[0])
        return res

    if len(label_cols) > 4:
        raise ValueError("Maximum of 4 labels can be specidied for analysis.")

    stop_words = set(stopwords.words(language))
    punctuations = set(string.punctuation)

    doc_lengths = []
    token_to_count = {}
    NNs = {}
    JJs = {}
    Vs = {}

    logger.info("Processing text in %s column of the input DataFrame..." % text_col)
    for text in tqdm(df[text_col]):
        try:
            tokens = text.lower().split(" ")
            doc_lengths.append(len(tokens))
            if skip_stopwords_punc:
                tokens = [t for t in tokens if t not in stop_words and t not in punctuations]
                update_count(token_to_count, tokens)

        except Exception as e:
            logger.warning("Processing entry --- %s --- lead to exception: %s" % (text, e.args[0]))
            continue

        postag_tokens = nltk.pos_tag(tokens)
        nouns = get_pos(postag_tokens, "NN")
        update_count(NNs, nouns)
        verbs = get_pos(postag_tokens, "VB")
        update_count(Vs, verbs)
        adjectives = get_pos(postag_tokens, "JJ")
        update_count(JJs, adjectives)

    word_frequencies = [v for _, v in token_to_count.items()]

    fig_main = create_adjust_subplots(label_cols)

    logger.info("Generating distplots and word cloud for input text")
    generate_text_plots(fig_main, doc_lengths, word_frequencies, NNs, JJs, Vs)
    logger.info("Generating plots for labels")
    generate_label_plots(fig_main, df, label_cols)
    logger.info("Rendering plots")
    fig_main.update_layout(height=3100, showlegend=False)
    if save_report:
        plotly.offline.plot(fig_main, filename=os.path.join(out_dir, "report.html"))
    else:
        fig_main.show()



def predict_dataframe(classification_model, data_frame):
            predictions = []
            for i in range(data_frame.shape[0]):
                predictions.append(classification_model.predict(data_frame.iloc[i]['text'])[0][0])
            groundtruth = data_frame['label'].tolist()
            return predictions, groundtruth


# -*- coding: utf-8 -*-

import os
from setuptools import setup, find_packages

if os.path.exists("README.md"):
    with open("README.md", "r") as f:
        long_description = f.read()
else:
    long_description = "Description could not be found."

if os.path.exists("requirements.txt"):
    with open("requirements.txt") as f:
        requirements = f.read().split("\n")
else:
    requirements = []

setup(
    name="snlp",
    version='0.1.3.post1',
    author="meghdadFar",
    author_email="meghdad.farahmand@gmail.com",
    description="Statistical NLP",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/meghdadFar/snlp",
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    scripts=['bin/downloads.py'],
    python_requires='>=3.7',
)


from typing import Dict, List
import math


def calculate_pmi(
    compound_dict: dict, word_dic: dict, num_compound: int, num_words: int, normalize: bool = False
) -> Dict[str, float]:
    """Calculate Pointwise Mutual Information between the two words of every word pair in nn_dict.

    Args:
        compound_dict: Dictionary of compounds and their count.
        word_dic: Dictionary of words and their count.
        num_compound: Number of compounds.
        num_words: Number of words.
        normalize: Whether or not normalize the pmi score. Normalized pmi is referred to as npmi.

    Returns:
        sorted_compound_dict: Dictionary of compounds and their pmi/npmi values, sorted wrt their pmi/npmi.
    """
    tmp_compound_dict = compound_dict
    for compound, count in tmp_compound_dict.items():
        w1w2 = compound.split(" ")
        # To filter out compounds that are rare/unique because of strange/misspelled component words.
        if float(word_dic[w1w2[0]]) > 10 and float(word_dic[w1w2[1]]) > 10:
            p_of_c = float(count) / float(num_words)
            p_of_h = float(word_dic[w1w2[0]]) / float(num_words)
            p_of_m = float(word_dic[w1w2[1]]) / float(num_words)
            pmi = math.log(p_of_c / (p_of_h * p_of_m))
            if not normalize:
                tmp_compound_dict[compound] = round(pmi, 2)
            else:
                npmi = pmi / float(-math.log(p_of_c))
                tmp_compound_dict[compound] = round(npmi, 2)
        else:
            tmp_compound_dict[compound] = 0.0
    sorted_compound_dict = dict(sorted(tmp_compound_dict.items(), key=lambda e: e[1], reverse=True))
    return sorted_compound_dict


def calculate_am(count_data: dict, am: str, mwe_types: List[str]) -> Dict[str, Dict]:
    """Read the counts from path_to_counts and for each compound calculates the measure specified by am.

    Args:
        count_data: A dictionary that contains different MWE types and their counts.
        am: Association measure to be used in order to extract MWEs. Can be any of [pmi, npmi]
        mwe_types: Types of MWEs. Can be any of [NC, JNC].

    Returns:
        res: Dictionary of MWE type to their individual MWE to its score dictionary.
    """
    res = {}
    num_words = sum(count_data["WORDS"].values())
    if am == "pmi":
        for mt in mwe_types:
            compound_dict_tmp = calculate_pmi(
                compound_dict=count_data[mt],
                word_dic=count_data["WORDS"],
                num_compound=sum(count_data[mt].values()),
                num_words=num_words,
                normalize=False,
            )
            res[mt] = compound_dict_tmp
    elif am == "npmi":
        for mt in mwe_types:
            compound_dict_tmp = calculate_pmi(
                compound_dict=count_data[mt],
                word_dic=count_data["WORDS"],
                num_compound=sum(count_data[mt].values()),
                num_words=num_words,
                normalize=True,
            )
            res[mt] = compound_dict_tmp
    return res
